package fr.labri.collections;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;

public abstract class AbstractTree implements ITree {

	protected ITree parent;
	protected List<ITree> children;

	@Override
	public int getChildPosition(ITree child) {
		return getChildren().indexOf(child);
	}

	@Override
	public ITree getChild(int position) {
		return getChildren().get(position);
	}

	@Override
	public List<ITree> getDescendants() {
		List<ITree> trees = TreeUtils.preOrder(this); 
		trees.remove(0);
		return trees;
	}

	@Override
	public boolean isIsomorphicTo(ITree other) {
		return TreeUtils.isIsomorphic(this, other);
	}
	
	@Override
	public boolean hasLabel() {
		return !ITree.NO_LABEL.equals(getLabel());
	}

	@Override
	public List<ITree> getLeaves() {
		List<ITree> leafs = new ArrayList<>();
		for (ITree t: getTrees()) if (t.isLeaf()) leafs.add(t);
		return leafs;
	}

	@Override
	public ITree getParent() {
		return parent;
	}

	@Override
	public void setParent(ITree parent) {
		this.parent = parent;
	}

	@Override
	public List<ITree> getParents() {
		List<ITree> parents = new ArrayList<>();
		if (getParent() == null) return parents;
		else {
			parents.add(getParent());
			parents.addAll(getParent().getParents());
		}
		return parents;
	}

	@Override
	public List<ITree> getTrees() {
		return TreeUtils.preOrder(this);
	}

	@Override
	public boolean isCompatible(ITree t) {
		return getType() == t.getType();
	}

	@Override
	public boolean isLeaf() {
		return getChildren().size() == 0;
	}

	@Override
	public boolean isRoot() {
		return getParent() == null;
	}

	@Override
	public boolean isSimilar(ITree t) {
		if (!isCompatible(t)) return false;
		else if (!getLabel().equals(t.getLabel())) return false;
		return true;
	}

	@Override
	public Iterable<ITree> postOrder() {
		return new Iterable<ITree>() {
			@Override
			public Iterator<ITree> iterator() {
				return TreeUtils.postOrderIterator(AbstractTree.this);
			}
		};
	}

	@Override
	public Iterable<ITree> breadthFirst() {
		return new Iterable<ITree>() {
			@Override
			public Iterator<ITree> iterator() {
				return TreeUtils.breadthFirstIterator(AbstractTree.this);
			}
		};
	}

	@Override
	public int positionInParent() {
		ITree p = getParent();
		if (p == null)
			return -1;
		else
			return p.getChildren().indexOf(this);
	}

	@Override
	public String toString() {
		throw new RuntimeException("This method should currently not be used (please use toShortString())");
	}


	public static class FakeTree extends AbstractTree {
		public FakeTree(ITree... trees) {
			children = new ArrayList<ITree>(trees.length);
			children.addAll(Arrays.asList(trees));
		}

		private RuntimeException unsupportedOperation() {
			return new UnsupportedOperationException("This method should not be called on a fake tree");
		}

		@Override
		public void addChild(ITree t) {
			throw unsupportedOperation();
		}

		@Override
		public ITree deepCopy() {
			throw unsupportedOperation();
		}

		@Override
		public List<ITree> getChildren() {
			return children;
		}

		@Override
		public String getLabel() {
			return Tree.NO_LABEL;
		}

		@Override
		public int getType() {
			return -1;
		}

		@Override
		public void setChildren(List<ITree> children) {
			throw unsupportedOperation();
		}

		@Override
		public void setLabel(String label) {
			throw unsupportedOperation();
		}

		@Override
		public void setParentAndUpdateChildren(ITree parent) {
			throw unsupportedOperation();
		}

		@Override
		public void setType(int type) {
			throw unsupportedOperation();
		}
	}
}
