package fr.labri.collections;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.NoSuchElementException;
import java.lang.Iterable;

public class PriorityQueue<E> implements Iterable<ArrayList<E>> {
	final int _maxPriority;
	final int _minPriority;
	final ArrayList<E>[] _groups;

	@SuppressWarnings("unchecked")
	public PriorityQueue(int min, int max) {
		_minPriority = min;
		_maxPriority = max;
		_groups = new ArrayList[_maxPriority - _minPriority + 1];
	}
	
	public static final <T> PriorityQueue<T> getPriorityQueue(int min, int max, final PriorityQueue.PriorityAdapter<T> adapter) {
		return new PriorityQueue<T>(min, max) {
			@Override
			public void insert(T item) {
				insert(adapter, item);
			}
		};
	}
	
	final public void insert(int priority, E item) {
		if (priority < _minPriority || priority > _maxPriority)
			throw new IndexOutOfBoundsException(String.format(
					"Priority %d not in range [%d..%d]: 0x%x", priority, _minPriority, _maxPriority, this.hashCode()));
		getGroup(priority - _minPriority).add(item);
	}

	public final void insert(PriorityAdapter<E> adapter, E item) {
		insert(adapter.getPriority(item), item);
	}

	public void insert(E item) {
		throw new UnsupportedOperationException("This PriorityQueue has no default PriorityAdapter");
	}

	@Override
	public PriorityIterator iterator() {
		return iterator(false);
	}

	public PriorityIterator iterator(boolean autoclean) {
		if (autoclean)
			return new PriorityIterator() {
				public ArrayList<E> next() {
					return pop();
				}
			};
		else
			return new PriorityIterator() {
				@Override
				public ArrayList<E> next() {
					return _next();
				}
			};
	}

	private ArrayList<E> getGroup(int pos) {
		ArrayList<E> group = _groups[pos];
		if (group == null)
			group = _groups[pos] = new ArrayList<>();
		return group;
	}

	private int getNextGroup(int pos) {
		while (pos > -1 && _groups[pos] == null)
			pos--;
		return pos;
	}

	public interface PriorityAdapter<E> {
		public int getPriority(E item);
	}

	public abstract class PriorityIterator implements Iterator<ArrayList<E>> {
		final int maxPriority = _maxPriority - _minPriority;
		int currentPriority = maxPriority;

		@Override
		public boolean hasNext() {
			return (currentPriority = getNextGroup(currentPriority)) != -1;
		}

		public int currentPriority() {
			return (currentPriority = getNextGroup(currentPriority)) + _minPriority;
		}

		public ArrayList<E> _next() {
			if (!hasNext())
				throw new NoSuchElementException();
			return getGroup(currentPriority--);
		}

		@Override
		public void remove() {
			if (currentPriority >= maxPriority)
				throw new IllegalStateException("Must call next before remove.");
			_groups[currentPriority + 1] = null;
		}

		public ArrayList<E> pop() {
			ArrayList<E> group = _next();
			remove();
			return group;
		}
	}
}