package fr.labri.collections;

import java.util.ArrayDeque;
import java.util.ArrayList;
import java.util.Deque;
import java.util.Iterator;
import java.util.List;
import java.util.NoSuchElementException;


public final class TreeUtils {

	// This class has no instances
	private TreeUtils() {}

	/**
	 * Compute the number of nodes in a tree (itself included).
	 * @param tree a Tree
	 */
	public static int getSize(ITree tree) {
		int size = 1;
		for (@SuppressWarnings("unused") ITree t: tree.postOrder()) {
			size ++;
		}
		return size;
	}

	/**
	 * Compute the depth of a node in a tree.
	 * @param tree a Tree
	 */
	public static int computeDepth(ITree tree) {
		int depth = 0;
		while (tree != null){
			depth ++;
			tree = tree.getParent();
		}
		return depth;
	}

	/**
	 * Compute the height of a node of the tree.
	 * @param tree a Tree.
	 */
	public static int computeHeight(ITree tree) {
		int height = 0;
		for (ITree t: tree.getChildren()) {
			int h = computeHeight(t);
			if (h > height)
				height = h;
		}
		return height + 1;
	}

	public static boolean isIsomorphic(ITree tree1, ITree tree2) {
		Iterator<ITree> it1 = postOrderIterator(tree1);
		Iterator<ITree> it2 = postOrderIterator(tree2);
		
		while (it1.hasNext() && it2.hasNext())
			if (!it1.next().isCompatible(it2.next()))
				return false;
		return !(it1.hasNext() || it2.hasNext());
	}
	/**
	 * Returns a list of every subtrees and the tree ordered using a pre-order.
	 * @param tree a Tree.
	 * @return
	 */
	public static List<ITree> preOrder(ITree tree) {
		List<ITree> trees = new ArrayList<>();
		preOrder(tree, trees);
		return trees;
	}

	private static void preOrder(ITree tree, List<ITree> trees) {
		// TODO the recursion should be externalized to avoid stack overflow
		trees.add(tree);
		if (!tree.isLeaf())
			for (ITree c: tree.getChildren())
				preOrder(c, trees);
	}

	/**
	 * Returns a list of every subtrees and the tree ordered using a breadth-first order.
	 * @param tree a Tree.
	 * @return
	 */
	public static List<ITree> breadthFirst(ITree tree) {
		List<ITree> trees = new ArrayList<>();
		List<ITree> currents = new ArrayList<>();
		currents.add(tree);
		while (currents.size() > 0) {
			ITree c = currents.remove(0);
			trees.add(c);
			currents.addAll(c.getChildren());
		}
		return trees;
	}
	
	/**
	 * Iterate on a tree and its descendent in bread-first order.
	 * @param tree a Tree.
	 * @return
	 */
	public static Iterator<ITree> breadthFirstIterator(final ITree tree) {
		return new Iterator<ITree>() {
			Deque<Iterator<ITree>> fifo = new ArrayDeque<>();
			
			{
				addLasts(new AbstractTree.FakeTree(tree));
			}
			
			@Override
			public boolean hasNext() {
				return !fifo.isEmpty();
			}
			
			@Override
			public ITree next() {
				while (!fifo.isEmpty()) {
					Iterator<ITree> it = fifo.getFirst();
					if (it.hasNext()) {
						ITree item = it.next();
						if (!it.hasNext())
							fifo.removeFirst();
						addLasts(item);
						return item;
					}
				}
				throw new NoSuchElementException();
			}
			private void addLasts(ITree item) {
				List<ITree> children = item.getChildren();
				if (!children.isEmpty())
					fifo.addLast(children.iterator());
			}
			@Override
			public void remove() {
				throw new RuntimeException("Not yet implemented implemented.");
			}
		};
	}

	/**
	 * Returns a list of every subtrees and the tree ordered using a post-order.
	 * @param tree a Tree.
	 * @return
	 */
	public static List<ITree> postOrder(ITree tree) {
		List<ITree> trees = new ArrayList<>();
		postOrder(tree, trees);
		return trees;
	}

	private static void postOrder(ITree tree, List<ITree> trees) {
		if (!tree.isLeaf()) for (ITree c: tree.getChildren()) postOrder(c, trees);
		trees.add(tree);
	}
	
	/**
	 * Iterate on a tree in post order, the first node is iterated last
	 * @param tree to iterator on
	 * @return
	 */
	public static Iterator<ITree> postOrderIterator(final ITree tree) {
		return new Iterator<ITree>() {
			Deque<Pair<ITree, Iterator<ITree>>> stack = new ArrayDeque<>();
			{
				push(tree);
			}
			
			@Override
			public boolean hasNext() {
				return stack.size() > 0;
			}

			@Override
			public ITree next() {
				if (stack.isEmpty())
					throw new NoSuchElementException();
				return selectNextChild(stack.peek().getSecond());
			}
			
			ITree selectNextChild(Iterator<ITree> it) {
				if (!it.hasNext())
					return stack.pop().getFirst();
				ITree item = it.next();
				if (item.isLeaf())
					return item;
				return selectNextChild(push(item));
			}

			private Iterator<ITree> push(ITree item) {
				Iterator<ITree> it = item.getChildren().iterator();
				stack.push(new Pair<>(item, it));
				return it;
			}

			@Override
			public void remove() {
				throw new RuntimeException("Not yet implemented implemented.");
			}
		};
	}
	
	/**
	 * Filter a Tree Iterator to only iterate on leaves 
	 * @param it a Tree Iterator
	 * @return a new iterator
	 */
	public static Iterator<ITree> leafIterator(final Iterator<ITree> it) {
		return new Iterator<ITree>() {
			ITree current = it.hasNext() ? it.next() : null;
			@Override
			public boolean hasNext() {
				return current != null;
			}

			@Override
			public ITree next() {
				ITree val = current;
				while (it.hasNext()) {
					current = it.next();
					if (current.isLeaf())
						break;
				}
				return val;
			}

			@Override
			public void remove() {
				throw new RuntimeException("Not yet implemented implemented.");
			}
		};
	}
}
